@@ -28,7 +28,7 @@
 #include <config.h>
 #include "gnome-vfs-parse-ls.h"
 
-#include "gnome-vfs-i18n.h"
+#include <glib/gi18n-lib.h>
 #include <glib/gmem.h>
 #include <glib/gmessages.h>
 #include <glib/gstrfuncs.h>
@@ -52,51 +52,56 @@
 #endif
 
 
-/* FIXME bugzilla.eazel.com 1179:
- * remove these globals.  */
-static int saveuid = -993;
-static char saveuname[TUNMLEN];
-static int my_uid = -993;
+typedef struct {
+	int saveuid;
+	int my_uid;
+	char saveuname[TUNMLEN];
+} finduid_cache;
+#define finduid_cache_init {-993, -993}
 
-static int savegid = -993;
-static char savegname[TGNMLEN];
-static int my_gid = -993;
+typedef struct {
+	int savegid;
+	int my_gid;
+	char savegname[TGNMLEN];
+} findgid_cache;
+#define findgid_cache_init {-993, -993}
 
-#define myuid	( my_uid < 0? (my_uid = getuid ()): my_uid )
-#define	mygid	( my_gid < 0? (my_gid = getgid ()): my_gid )
-
-static int finduid (char *uname)
+static int finduid (char *uname, finduid_cache *cache)
 {
 	struct passwd *pw;
 	
-	if (uname[0] != saveuname[0]/* Quick test w/o proc call */
-	    || 0 != strncmp (uname, saveuname, TUNMLEN)) {
-		strncpy (saveuname, uname, TUNMLEN);
+	if (uname[0] != cache->saveuname[0]/* Quick test w/o proc call */
+	    || 0 != strncmp (uname, cache->saveuname, TUNMLEN)) {
+		strncpy (cache->saveuname, uname, TUNMLEN);
 		pw = getpwnam (uname);
 		if (pw) {
-			saveuid = pw->pw_uid;
+			cache->saveuid = pw->pw_uid;
 		} else {
-			saveuid = myuid;
+			if (cache->my_uid < 0)
+				cache->my_uid = getuid ();
+			cache->saveuid = cache->my_uid;
 		}
 	}
-	return saveuid;
+	return cache->saveuid;
 }
 
-static int findgid (char *gname)
+static int findgid (char *gname, findgid_cache *cache)
 {
 	struct group *gr;
 	
-	if (gname[0] != savegname[0]/* Quick test w/o proc call */
-	    || 0 != strncmp (gname, savegname, TUNMLEN)) {
-		strncpy (savegname, gname, TUNMLEN);
+	if (gname[0] != cache->savegname[0]/* Quick test w/o proc call */
+	    || 0 != strncmp (gname, cache->savegname, TUNMLEN)) {
+		strncpy (cache->savegname, gname, TUNMLEN);
 		gr = getgrnam (gname);
 		if (gr) {
-			savegid = gr->gr_gid;
+			cache->savegid = gr->gr_gid;
 		} else {
-			savegid = mygid;
+			if (cache->my_gid < 0)
+				cache->my_gid = getgid ();
+			cache->savegid = cache->my_gid;
 		}
 	}
-	return savegid;
+	return cache->savegid;
 }
 
 
@@ -355,16 +360,35 @@ vfs_parse_filemode (const char *p)
 	return res;
 }
 
+static gboolean
+is_last_column (int idx,
+		int num_cols,
+		const char *carbon,
+		int column_ptr[])
+{
+	const char *p;
+
+	if (idx + 1 == num_cols) {
+		return TRUE;
+	}
+
+	p = carbon + column_ptr[idx + 1] - 1;
+	return *p == '\r' || *p == '\n';
+}
+
 static int
 vfs_parse_filedate (int idx,
 		    char *columns[],
+		    int num_cols,
+		    const char *carbon,
+		    int column_ptr[],
 		    time_t *t)
 {	/* This thing parses from idx in columns[] array */
 
 	char *p;
 	struct tm tim;
 	int d[3];
-	int	got_year = 0;
+	int got_year = 0, got_time = 0;
 	int current_mon;
 	time_t now;
     
@@ -434,14 +458,20 @@ vfs_parse_filedate (int idx,
 	}
 
 	/* Here we expect to find time and/or year */
-    
+
 	if (is_num (columns[idx])) {
-		if (is_time (columns[idx], &tim) || (got_year = is_year (columns[idx], &tim))) {
+		if ((got_time = is_time (columns[idx], &tim)) ||
+		    (got_year = is_year (columns[idx], &tim))) {
 			idx++;
 
 			/* This is a special case for ctime () or Mon DD YYYY hh:mm */
 			if (is_num (columns[idx]) && 
-			   ((got_year = is_year (columns[idx], &tim)) || is_time (columns[idx], &tim)))
+			    !is_last_column (idx, num_cols, carbon, column_ptr) && /* ensure that we don't eat two lines at once,
+										      where the first line provides a year-like
+										      filename but no year, or a time-like filename
+										      but no time */
+			    ((!got_year && (got_year = is_year (columns[idx], &tim))) ||
+			     (!got_time && (got_time = is_time (columns[idx], &tim)))))
 				idx++; /* time & year or reverse */
 		} /* only time or date */
 	}
@@ -477,11 +507,14 @@ gnome_vfs_parse_ls_lga (const char *p,
 	int i;
 	int nlink;
 	char *p_copy, *p_pristine;
+	finduid_cache uid_cache = finduid_cache_init;
+	findgid_cache gid_cache = findgid_cache_init;
 
 	if (strncmp (p, "total", 5) == 0)
 		return 0;
 
 	p_copy = g_strdup (p);
+
 	if ((i = vfs_parse_filetype (*(p++))) == -1)
 		goto error;
 
@@ -521,7 +554,7 @@ gnome_vfs_parse_ls_lga (const char *p,
 	s->st_nlink = nlink;
 
 	if (!is_num (columns[1]))
-		s->st_uid = finduid (columns [1]);
+		s->st_uid = finduid (columns [1], &uid_cache);
 	else
 		s->st_uid = (uid_t) atol (columns [1]);
 
@@ -546,7 +579,7 @@ gnome_vfs_parse_ls_lga (const char *p,
 		if (is_num (columns[2]))
 			s->st_gid = (gid_t) atol (columns [2]);
 		else
-			s->st_gid = findgid (columns [2]);
+			s->st_gid = findgid (columns [2], &gid_cache);
 		idx2 = 3;
 	}
 
@@ -585,7 +618,7 @@ gnome_vfs_parse_ls_lga (const char *p,
 #endif
 	}
 
-	idx = vfs_parse_filedate (idx, columns, &s->st_mtime);
+	idx = vfs_parse_filedate (idx, columns, num_cols, p, column_ptr, &s->st_mtime);
 	if (!idx)
 		goto error;
 	/* Use resulting time value */
